#pragma once


namespace Robotics
{
    ref class Vector;


    /// s
    public ref class Matrix
    {
    // Construction --------------------------------------------------------
    public:
        Matrix( Matrix^ other );
        Matrix( int row, int column );


    // Operation -----------------------------------------------------------
    public:
        void SetSize( int row, int column );

        void UnitMatrix();
        void ZeroMatrix();

        Matrix^ TransposedMatrix();
        bool InverseMatrix( Matrix^% result );

        bool IsUnitMatrix( double precision );
        bool IsZeroMatrix();
        bool IsOrthogonalMatrix( double precision );

        virtual System::String^ ToString() override;

    private:
        void SecureDomain( int row );

        bool LuDecomposition( array< int >^% informationToChangeRow );
        array< double >^ GetComponentWithMaximumAbsoluteValue();

        void RemoveErrorComponent();

        bool IsEqualSize( Matrix^ obj );
        bool IsSquareMatrix();


    // Overload ------------------------------------------------------------
    public:
        Matrix^ operator+=( Matrix^ obj );
        Matrix^ operator-=( Matrix^ obj );
        Matrix^ operator*=( double val );
        Matrix^ operator*=( Matrix^ obj );

        virtual bool Equals( Matrix^ obj );
        virtual bool Equals( Matrix^ obj, double precision );

        static bool operator==( Matrix^ left, Matrix^ right );
        static bool operator!=( Matrix^ left, Matrix^ right );

        static Matrix^ operator+( Matrix^ left, Matrix^ right );
        static Matrix^ operator-( Matrix^ left, Matrix^ right );

        static Vector^ operator*( Matrix^ matrix, Vector^ vector );
        static Vector^ operator*( Vector^ vector, Matrix^ matrix );
        static Matrix^ operator*( Matrix^ left, Matrix^ right );


        /// @note
        /// CX^Xƃnh鋰ꂪ邽߁A
        /// Z͑d`ĂȂB


    // Attribute ===========================================================
    private:
        array< Vector^ >^ m_rowVectors;    ///< sxNg


    // static
        literal double Tolerance = 1.0E-8;    ///< e덷


    // Property ------------------------------------------------------------
    public:
        /// Indexer
        property Vector^ default[ int ]
        {
            Vector^ get( int index );
        }

        /// s
        property int Row
        {
            int get()
            {
                return m_rowVectors->Length;
            }
        }

        /// 
        property int Column
        {
            int get();
        }
    };
}

// [ Reference ]ulvZȑOvYamada.K ( http://www.asahi-net.or.jp/~uc3k-ymd/index.html )