#include "RotationMatrix.h"


#include "AxisVector.h"


using namespace Robotics;
using namespace System;
using namespace System::Diagnostics;


/// [ Constructor ]
RotationMatrix::RotationMatrix() :
    Matrix( DefaultRow, DefaultColumn )
{
    // Pʍsɏ
    UnitMatrix();

    Debug::Assert( IsOrthogonalMatrix(), "sł" );
}

/// [Esb`E[p琶 [ Constructor ]
RotationMatrix::RotationMatrix( double roll, double pitch, double yaw ) :
    Matrix( DefaultRow, DefaultColumn )
{
    Debug::Assert( Angle::IsExpressedByRadian( roll ), "Radianŕ\Ă" );
    Debug::Assert( Angle::IsExpressedByRadian( pitch ), "Radianŕ\Ă" );
    Debug::Assert( Angle::IsExpressedByRadian( yaw ), "Radianŕ\Ă" );


    // [Esb`E[p ꂼ̃TC߂
    double sinRoll = Math::Sin( roll );
    double sinPitch = Math::Sin( pitch );
    double sinYaw = Math::Sin( yaw );

    // ... RTC߂
    double cosRoll = Math::Cos( roll );
    double cosPitch = Math::Cos( pitch );
    double cosYaw = Math::Cos( yaw );


    this[ 1 ][ 1 ] = cosYaw * cosPitch;
    this[ 2 ][ 1 ] = sinYaw * cosPitch;
    this[ 3 ][ 1 ] = -sinPitch;

    this[ 1 ][ 2 ] = ( -sinYaw * cosRoll ) + ( cosYaw * sinPitch * sinRoll );
    this[ 2 ][ 2 ] = ( cosYaw * cosRoll ) + ( sinYaw * sinPitch * sinRoll );
    this[ 3 ][ 2 ] = cosPitch * sinRoll;

    this[ 1 ][ 3 ] = ( sinYaw * sinRoll ) + ( cosYaw * sinPitch * cosRoll );
    this[ 2 ][ 3 ] = ( -cosYaw * sinRoll ) + ( sinYaw * sinPitch * cosRoll );
    this[ 3 ][ 3 ] = cosPitch * cosRoll;


    Debug::Assert( IsOrthogonalMatrix(), "sł" );

// [ Reference ]uq[}mCh{bgvcGi(I[) P25
// [ Reference ]um肽{bgn݌vvēc(uk) P67
}

/// s񂩂琶 [ Constructor ]
RotationMatrix::RotationMatrix( Matrix^ matrix ) :
    Matrix( matrix )
{
    Debug::Assert( IsCorrectSize( matrix ), "s̑傫͐" );
    Debug::Assert( IsOrthogonalMatrix(), "sł" );
}

/// XML琶 [ Constructor ]
RotationMatrix::RotationMatrix( System::Xml::XmlElement^ element ) :
    Matrix( DefaultRow, DefaultColumn )
{
    // Pʍsɏ
    UnitMatrix();


    // vf擾
    Xml::XmlElement^ node = element[ "rotationMatrix" ];

    if( node != nullptr )
    {
        // 擾łȂ΁Aݒ肷
        this[ 1 ][ 1 ] = Double::Parse( node->GetAttribute( "c11" ) );
        this[ 1 ][ 2 ] = Double::Parse( node->GetAttribute( "c12" ) );
        this[ 1 ][ 3 ] = Double::Parse( node->GetAttribute( "c13" ) );

        this[ 2 ][ 1 ] = Double::Parse( node->GetAttribute( "c21" ) );
        this[ 2 ][ 2 ] = Double::Parse( node->GetAttribute( "c22" ) );
        this[ 2 ][ 3 ] = Double::Parse( node->GetAttribute( "c23" ) );

        this[ 3 ][ 1 ] = Double::Parse( node->GetAttribute( "c31" ) );
        this[ 3 ][ 2 ] = Double::Parse( node->GetAttribute( "c32" ) );
        this[ 3 ][ 3 ] = Double::Parse( node->GetAttribute( "c33" ) );
    }
    Debug::Assert( IsOrthogonalMatrix(), "sł" );
}


///  * [ Operator Overloading ]
RotationMatrix^ RotationMatrix::operator*( RotationMatrix^ left, RotationMatrix^ right )
{
    RotationMatrix^ result = gcnew RotationMatrix( left );
    result *= right;

    return result;
}

///  * [ Operator Overloading ]
RotationMatrix^ RotationMatrix::operator*( RotationMatrix^ rotationMatrix, Matrix^ matrix )
{
    Debug::Assert( IsCorrectSize( matrix ), "s̑傫͐" );

    RotationMatrix^ result = gcnew RotationMatrix( rotationMatrix );
    result *= matrix;

    return result;
}

///  * [ Operator Overloading ]
RotationMatrix^ RotationMatrix::operator*( Matrix^ matrix, RotationMatrix^ rotationMatrix )
{
    Debug::Assert( IsCorrectSize( matrix ), "s̑傫͐" );

    RotationMatrix^ result = gcnew RotationMatrix( matrix );
    result *= rotationMatrix;

    return result;
}


/// l̓mF
bool RotationMatrix::Equals( RotationMatrix^ obj, Angle precision )
{
    // NULLƂ͓ł͂Ȃ
    if( safe_cast< System::Object^ >( obj ) == nullptr ) return false;


    // [pAsb`pA[pꂼɂāAxw肵Ĕr
    return ( this->Roll.Equals( obj->Roll, precision ) )
        && ( this->Pitch.Equals( obj->Pitch, precision ) )
        && ( this->Yaw.Equals( obj->Yaw, precision ) );
}


/// w֐Ԃ (hQX̌ / pxxNgɂĐ]s߂)
RotationMatrix^ RotationMatrix::MatrixExponential( Vector^ axisVector, Angle angle )
{
    // ЂݑΏ̍s(s)߂
    Matrix^ skewSymmetricMatrix = axisVector->SkewSymmetricMatrix();

    // ... ̂Q߂
    Matrix^ squareSkewSymmetricMatrix = skewSymmetricMatrix * skewSymmetricMatrix;

    // ... āAꂼ̍sXJ[{
    skewSymmetricMatrix *= Math::Sin( angle );
    squareSkewSymmetricMatrix *= 1.0 - Math::Cos( angle );


    RotationMatrix^ result = gcnew RotationMatrix();
    Debug::Assert( result->IsUnitMatrix(), "]s͒PʍsŏĂ" );

    // ... ɋ߂Q̍sZ
    result += skewSymmetricMatrix + squareSkewSymmetricMatrix;

    return result;

// [ Reference ]uq[}mCh{bgvcGi(I[) P34
}


/// sΐԂ
Matrix^ RotationMatrix::LogMatrix()
{
    // pxxNg߂
    Vector^ angularVelocity = AngularVelocityVector();

    // ... ̂ЂݑΏ̍sԂ
    return angularVelocity->SkewSymmetricMatrix();


// [ Reference ]uq[}mCh{bgvcGi(I[) P35
}

/// pxxNgԂ (PbԂ Ή]sɓBl)
Vector^ RotationMatrix::AngularVelocityVector()
{
    if( IsUnitMatrix() )
    {
        // PʍsȂ xNgԂ
        Vector^ result = gcnew Vector( 3 );
        Debug::Assert( result->IsZeroVector(), "xNgŏĂ" );

        return result;
    }


    double parameterOfArcCosine
        = ( this[ 1 ][ 1 ] + this[ 2 ][ 2 ] + this[ 3 ][ 3 ] - 1.0 ) / 2.0;

    Debug::Assert(
        ( -1.0 <= parameterOfArcCosine ) &&
        ( +1.0 >= parameterOfArcCosine ),
        "t]̈͐ł"
    );
    double theta = Math::Acos( parameterOfArcCosine );
    double scalar = theta / ( 2.0 * Math::Sin( theta ) );


    Vector^ result = gcnew Vector( 3 );
    result[ 1 ] = scalar * ( this[ 3 ][ 2 ] - this[ 2 ][ 3 ] );
    result[ 2 ] = scalar * ( this[ 1 ][ 3 ] - this[ 3 ][ 1 ] );
    result[ 3 ] = scalar * ( this[ 2 ][ 1 ] - this[ 1 ][ 2 ] );

    return result;


// [ Reference ]uq[}mCh{bgvcGi(I[) P35
}


/// X̎xNg擾 [ Property ]
AxisVector^ RotationMatrix::X::get()
{
    AxisVector^ result = gcnew AxisVector(
        this[ 1 ][ 1 ],
        this[ 2 ][ 1 ],
        this[ 3 ][ 1 ]
    );
    return result;
}

/// Y̎xNg擾 [ Property ]
AxisVector^ RotationMatrix::Y::get()
{
    AxisVector^ result = gcnew AxisVector(
        this[ 1 ][ 2 ],
        this[ 2 ][ 2 ],
        this[ 3 ][ 2 ]
    );
    return result;
}

/// Z̎xNg擾 [ Property ]
AxisVector^ RotationMatrix::Z::get()
{
    AxisVector^ result = gcnew AxisVector(
        this[ 1 ][ 3 ],
        this[ 2 ][ 3 ],
        this[ 3 ][ 3 ]
    );
    return result;
}


/// [p擾 [ Property ]
Angle RotationMatrix::Roll::get()
{
    double y = this[ 3 ][ 2 ];
    double x = this[ 3 ][ 3 ];

    return Angle( Atan2( y, x ) );

// [ Reference ]um肽{bgn݌vvēc(uk) P67
}

/// sb`p擾 [ Property ]
Angle RotationMatrix::Pitch::get()
{
    double y = this[ 3 ][ 1 ];
    double x = Math::Sqrt(
        this[ 1 ][ 1 ] * this[ 1 ][ 1 ] +
        this[ 2 ][ 1 ] * this[ 2 ][ 1 ]
    );

    return Angle( Atan2( -y, x ) );

// [ Reference ]um肽{bgn݌vvēc(uk) P67
}

/// [p擾 [ Property ]
Angle RotationMatrix::Yaw::get()
{
    double y = this[ 2 ][ 1 ];
    double x = this[ 1 ][ 1 ];

    return Angle( Atan2( y, x ) );

// [ Reference ]um肽{bgn݌vvēc(uk) P67
}


/// ɋ߂ [p擾
Angle RotationMatrix::GetRoll( Angle standard )
{
    double y = this[ 3 ][ 2 ];
    double x = this[ 3 ][ 3 ];


    // ςāAQ̉߂
    Angle result1( Atan2( +y, +x ) );
    Angle result2( Atan2( -y, -x ) );

    // ... ɋ߂Ԃ
    return standard.Near( result1, result2 );
}

/// ɋ߂ sb`p擾
Angle RotationMatrix::GetPitch( Angle standard )
{
    double y = this[ 3 ][ 1 ];
    double x = Math::Sqrt(
        this[ 1 ][ 1 ] * this[ 1 ][ 1 ] +
        this[ 2 ][ 1 ] * this[ 2 ][ 1 ]
    );


    // ςāAQ̉߂
    Angle result1( Atan2( -y, +x ) );
    Angle result2( Atan2( -y, -x ) );

    // ... ɋ߂Ԃ
    return standard.Near( result1, result2 );
}

/// ɋ߂ [p擾
Angle RotationMatrix::GetYaw( Angle standard )
{
    double y = this[ 2 ][ 1 ];
    double x = this[ 1 ][ 1 ];


    // ςāAQ̉߂
    Angle result1( Atan2( +y, +x ) );
    Angle result2( Atan2( -y, -x ) );

    // ... ɋ߂Ԃ
    return standard.Near( result1, result2 );
}


/// 傫͐H
bool RotationMatrix::IsCorrectSize( Matrix^ obj )
{
    // s͊̍słA񐔂̗͊񐔂H
    return ( obj->Row == DefaultRow )
        && ( obj->Column == DefaultColumn );
}

/// H
bool RotationMatrix::IsHorizontal( double precision )
{
    // [pƃsb`p [ł邱ƂmF
    return ( Math::Abs( Roll ) < precision )
        && ( Math::Abs( Pitch ) < precision );
}

/// ɂ
void RotationMatrix::Horizontal()
{
    double sinYaw = Math::Sin( Yaw );
    double cosYaw = Math::Cos( Yaw );

    this[ 1 ][ 1 ] = cosYaw;
    this[ 2 ][ 1 ] = sinYaw;
    this[ 3 ][ 1 ] = 0.0;

    this[ 1 ][ 2 ] = -sinYaw;
    this[ 2 ][ 2 ] = cosYaw;
    this[ 3 ][ 2 ] = 0.0;

    this[ 1 ][ 3 ] = 0.0;
    this[ 2 ][ 3 ] = 0.0;
    this[ 3 ][ 3 ] = 1.0;


    Debug::Assert( IsHorizontal( Double::Epsilon ), "ł" );
}


/// ^WFg 2̎w肳ꂽ̏łpxԂ
double RotationMatrix::Atan2( double y, double x )
{
    if( x == 0 && y == 0 )
    {
        // 2̈[̏ꍇ́A[Ԃ
        return 0.0;
    }
    else
    {
        // MathNX ֐Ăяo
        return Math::Atan2( y, x );
    }


    /// @note
    /// 2̈[̏ꍇ̖߂nɂāAC̕WCuł͏n`ƂĂB
    /// Ƃ낪 .NET MathNXɂẮA߂lLĂȂB
    /// āAŏ𖾎IɍsĂB
    /// http://msdn2.microsoft.com/ja-jp/library/system.math.atan2.aspx
}


/// CX^X̐𕶎ŕԂ (MATLAB`)
System::String^ RotationMatrix::ToString()
{
    return String::Format(
        "[ {0:0.0000000} {1:0.0000000} {2:0.0000000};" + Environment::NewLine +
        "  {3:0.0000000} {4:0.0000000} {5:0.0000000};" + Environment::NewLine +
        "  {6:0.0000000} {7:0.0000000} {8:0.0000000} ]",
        this[ 1 ][ 1 ], this[ 1 ][ 2 ], this[ 1 ][ 3 ],
        this[ 2 ][ 1 ], this[ 2 ][ 2 ], this[ 2 ][ 3 ],
        this[ 3 ][ 1 ], this[ 3 ][ 2 ], this[ 3 ][ 3 ]
    );
}

/// CX^X̐[Esb`E[\̕ŕԂ
System::String^ RotationMatrix::ToStringByRollPitchYaw()
{
    return String::Format(
        "Roll:{0:0.0} Pitch:{1:0.0} Yaw:{2:0.0}",
        Roll.Degree, Pitch.Degree, Yaw.Degree
    );
}


/// CX^X̐XMLŕԂ
System::Xml::XmlElement^ RotationMatrix::ToXml( System::Xml::XmlDocument^ document )
{
    // [g𐶐
    Xml::XmlElement^ root = document->CreateElement( "rotationMatrix" );

    // ... tB[h𑮐ƂĐݒ肷
    root->SetAttribute( "c11", this[ 1 ][ 1 ].ToString() );
    root->SetAttribute( "c12", this[ 1 ][ 2 ].ToString() );
    root->SetAttribute( "c13", this[ 1 ][ 3 ].ToString() );

    root->SetAttribute( "c21", this[ 2 ][ 1 ].ToString() );
    root->SetAttribute( "c22", this[ 2 ][ 2 ].ToString() );
    root->SetAttribute( "c23", this[ 2 ][ 3 ].ToString() );

    root->SetAttribute( "c31", this[ 3 ][ 1 ].ToString() );
    root->SetAttribute( "c32", this[ 3 ][ 2 ].ToString() );
    root->SetAttribute( "c33", this[ 3 ][ 3 ].ToString() );

    return root;
}
