#pragma once


namespace Robotics
{
    ref class Matrix;


    /// xNg
    public ref class Vector
    {
    // Construction --------------------------------------------------------
    public:
        Vector( Vector^ other );

        explicit Vector( int dimension );
        Vector( const double* val, int dimension );


    // Operation -----------------------------------------------------------
    public:
        void ZeroVector();
        double Norm();

        Matrix^ SkewSymmetricMatrix();

        bool IsUnitVector( double precision );
        bool IsZeroVector();

        void RemoveErrorComponent();

        virtual System::String^ ToString() override;

    private:
        void SecureDomain( int dimension );
        bool IsRangeOfDimension( int index );


    // static
    public:
        static double InnerProduct( Vector^ left, Vector^ right );
        static Vector^ CrossProduct( Vector^ left, Vector^ right );
        static Vector^ Combine( Vector^ left, Vector^ right );


    // Overload ------------------------------------------------------------
    public:
        Vector^ operator-();

        Vector^ operator*=( double val );
        Vector^ operator/=( double val );
        Vector^ operator+=( Vector^ obj );
        Vector^ operator-=( Vector^ obj );

        virtual bool Equals( Vector^ obj );
        virtual bool Equals( Vector^ obj, double precision );

        static bool operator==( Vector^ left, Vector^ right );
        static bool operator!=( Vector^ left, Vector^ right );

        static Vector^ operator+( Vector^ left, Vector^ right );
        static Vector^ operator-( Vector^ left, Vector^ right );

        static Vector^ operator*( double scalar, Vector^ obj );
        static Vector^ operator*( Vector^ obj, double scalar );
        static Vector^ operator/( Vector^ obj, double scalar );


        /// @note
        /// CX^Xƃnh鋰ꂪ邽߁A
        /// Z͑d`ĂȂB


    // Attribute ===========================================================
    private:
        array< double >^ m_components;  ///< 


    // static
        literal double Tolerance = 1.0E-8;    ///< e덷


    // Property ------------------------------------------------------------
    public:
        /// Indexer
        property double default[ int ]
        {
            double get( int index );
            void set( int index, double value );
        }

        /// 
        [ System::ComponentModel::Browsable( false ) ]  // Qƕs
        property int Dimension
        {
            int get()
            {
                return m_components->Length;
            }
            void set( int value );
        }
    };
}

// [ Reference ]ulvZȑOvYamada.K ( http://www.asahi-net.or.jp/~uc3k-ymd/index.html )